#include <OS.h>

#include <vector>

#include <getopt.h>
#include <stdio.h>
#include <unistd.h>

typedef vector<area_info> area_t;

void DisplayUsage(void) {
	const char *options[][4] = {
		{ "help", "h", "Display this message", "none" },
		{ "name", "a", "Displays only areas matching name", "string" },
		{ "id", "i", "Show only the area matching the id", "integer" },
		{ "output", "o", "File to dump memory contents to (overwrites file, defaults to smithy.dump)", "string" },
		{ NULL, NULL, NULL, NULL }
	};

	printf("Valid options are;\n");
	
	for (int32 i = 0; options[i][0] != NULL; i++) {
		printf("  --%s (%s)\t%s (Argument: %s)\n", options[i][0], options[i][1],
			options[i][2], options[i][3]);

	};
};

status_t MatchArea(const char *areaName, area_info *info) {
	if (areaName != NULL) {
		area_id quickCheck = find_area(areaName);
		if (quickCheck == B_NAME_NOT_FOUND) {
			printf("No area named \"%s\"\n", areaName);
			return quickCheck;
		};
	};
		
	team_info teamInfo;
	int32 teamCookie = 0;
	area_t areas;
	
	while (get_next_team_info(&teamCookie, &teamInfo) == B_OK) {
		area_info areaInfo;
		int32 areaCookie = 0;
		bool hasArea = false;
		
		while (get_next_area_info(teamInfo.team, &areaCookie, &areaInfo) == B_OK) {
			if ((areaName == NULL) || (strcmp(areaInfo.name, areaName) == 0)) {
				if (hasArea == false) printf("%s\n", teamInfo.args);
				hasArea = true;
				
				printf("\t%i: %s (%i) (%i bytes @ [%p])\n", areas.size() + 1, areaInfo.name,
					areaInfo.area, areaInfo.size, areaInfo.address);
				areas.push_back(areaInfo);
			};
		};
	};

	area_info areaInfo;

	if (areas.size() > 1) {
		int32 index = -1;
		while ((index < 1) || (index > areas.size())) {
			printf("Please enter the number of the area you wish to dump (%i - %i): ", 1,
				areas.size());
			scanf("%i", &index);
			printf("\n");
		};
		
		*info = areas[index];
	} else {
		*info = areas[0];
	};		

	return B_OK;
};

int main(int argc, char *argv[]) {
	char *areaMatchName = NULL;
	char *outputPath = "smithy.dump";
	area_id areaMatchId = -1;
	area_info dumpArea;
	
	struct option options[] = {
		{"help", false, NULL, 'h'},
		{"name", true, NULL, 'n'},
		{"id", true, NULL, 'i'},
		{"output", true, NULL, 'o'},
		{NULL, false, NULL, 0}
	};

	bool moreArgs = true;
	int optionIndex = 0;

	while (moreArgs == true) {
		int option = getopt_long(argc, argv, "hn:i:o:", options, &optionIndex);
					
		switch (option) {
			case 'h': 
			case '?': {
				DisplayUsage();
				return B_ERROR;;
			} break;
			case ':': {
				DisplayUsage();
				printf("%s requires a parameter\n", options[optind].name);
				return B_ERROR;
			} break;
			case -1: {
				moreArgs = false;
			} break;
			
			case 'n': areaMatchName = optarg; break;
			case 'i': areaMatchId = atoi(optarg); break;
			case 'o': outputPath = optarg; break;
		};
	};
	if (optind > argc) {
		DisplayUsage();
		return B_ERROR;
	};

	if (areaMatchId > -1) {
		status_t result = get_area_info(areaMatchId, &dumpArea);
		if (result != B_OK) {
			printf("Unable to locate area_id %i: %s\n", areaMatchId, strerror(result));
			return result;
		};
	} else {
		if (MatchArea(areaMatchName, &dumpArea) != B_OK) {
			printf("Unable to match area name.\n");
			return B_ERROR;
		};
	};

	void *addr = NULL;
	area_id clone = clone_area("smithy_clone", &addr, B_ANY_ADDRESS, B_READ_AREA, dumpArea.area);
	if (clone < B_OK) {
		printf("Unable to access address: %s\n", strerror(clone));
		return clone;
	};

	FILE *out = fopen(outputPath, "wb");
	fwrite(addr, dumpArea.size, sizeof(uchar), out);
	fclose(out);

	delete_area(clone);

	return B_OK;
};
